import os.path
import sys
import traceback

import requests

from file_list import ReleaseFile
from util import retry_multi, GLOBAL_TIMEOUT

LINUX_KEY = "Linux"
MACOSX_X64_KEY = "Mac-x86_64"
MACOSX_ARM64_KEY = "Mac-arm64"
WIN32_SSE2_KEY = "Win32-SSE2"
WIN64_SSE2_KEY = "Win64-SSE2"
WIN32_AVX_KEY = "Win32-AVX"
WIN64_AVX_KEY = "Win64-AVX"

metadata = {
    'type': 'engine',
    'title': 'FSO',
    'notes': '',
    'banner': 'https://fsnebula.org/storage/0d/e7/bf64bcdea9a9c115969cfb784e1ca457d24a7c2da4fc6f213521c3bb6abb.png',
    'screenshots': [],
    'videos': [],
    'first_release': '2017-09-24',
    'cmdline': '',
    'mod_flag': ['FSO'],
    'tile': 'https://fsnebula.org/storage/ec/cc/0bf23e028c26d5175ff52d003bff85b0a17b0ddfc1130d65bdf6d36f6324.png',
    'logo': None,
    'release_thread': None,
    'description': '',
    'id': 'FSO',
    'packages': [],
}

platforms = {
    LINUX_KEY: 'linux',
    MACOSX_X64_KEY: 'macosx',
    MACOSX_ARM64_KEY: 'macosx',
    WIN32_SSE2_KEY: 'windows',
    WIN64_SSE2_KEY: 'windows',
    WIN32_AVX_KEY: 'windows',
    WIN64_AVX_KEY: 'windows'
}

envs = {
    LINUX_KEY: 'linux && x86_64',  # Linux only has 64bit builds
    MACOSX_X64_KEY: 'macosx && x86_64',
    MACOSX_ARM64_KEY: 'macosx && arm64',
    WIN32_SSE2_KEY: 'windows',
    WIN64_SSE2_KEY: 'windows && x86_64',
    WIN32_AVX_KEY: 'windows && avx',
    WIN64_AVX_KEY: 'windows && avx && x86_64'
}

subdirs = {
    WIN32_SSE2_KEY: 'x86',
    WIN64_SSE2_KEY: 'x64',
    WIN32_AVX_KEY: 'x86_avx',
    WIN64_AVX_KEY: 'x64_avx'
}


def render_nebula_release(version, stability, files, config):
    meta = metadata.copy()
    meta['version'] = str(version)
    meta['stability'] = stability  # This can be one of ('stable', 'rc', 'nightly')

    for file in files:
        if file.content_hashes is None:
            # The extraction probably failed which is why we don't have any hashes
            continue

        group = file.group

        # release.py sets subgroup but nightly.py doesn't which means we have to normalize group here.
        if file.subgroup:
            group += '-' + file.subgroup

        pkg = {
            'name': group,
            'notes': '',
            'is_vp': False,
            'files': [{
                'dest': platforms[group] + '/' + subdirs.get(group, ''),
                'filesize': file.size,
                'checksum': ['sha256', file.hash],
                'urls': [file.url] + file.mirrors,
                'filename': file.name
            }],
            'environment': envs.get(group),
            'filelist': [],
            'status': 'required',
            'folder': None,
            'dependencies': [],
            'executables': []
        }

        for fn, checksum in file.content_hashes:
            if group in subdirs:
                dest_fn = subdirs[group] + '/' + fn
            else:
                dest_fn = fn

            # If the group is a known platform then we put the binaries into a separate subfolder for each platform
            if group in platforms:
                dest_fn = platforms[group] + '/' + dest_fn

            pkg['filelist'].append({
                'orig_name': fn,
                'checksum': ('sha256', checksum),
                'archive': file.name,
                'filename': dest_fn
            })

            if group == LINUX_KEY and fn.endswith('.AppImage'):
                if 'qtfred' in fn:
                    if '-FASTDBG' in fn:
                        label = 'QtFRED Debug'
                    else:
                        label = 'QtFRED'

                else:
                    if '-FASTDBG' in fn:
                        label = 'Fast Debug'
                    else:
                        label = None

                props = {
                    "x64": True,  # All Linux builds are 64-bit
                    "sse2": True,  # Linux builds are forced to compile with SSE2 but not AVX
                    "avx": False,
                    "avx2": False,
                }

                pkg['executables'].append({
                    'file': dest_fn,
                    'label': label,
                    'properties': props,
                })
            elif group.startswith('Mac') and fn.startswith(os.path.basename(fn) + '.app/'):
                if 'qtfred' in fn:
                    if '-FASTDBG' in fn:
                        label = 'QtFRED Debug'
                    else:
                        label = 'QtFRED'
                else:
                    if '-FASTDBG' in fn:
                        label = 'Fast Debug'
                    else:
                        label = None

                props = {
                    "arm64": "arm64" in fn,
                    "x64": "x64" in fn,
                    "sse2": True,  # There are no forced options on mac but 64-bit always implies SSE2 so we set that here
                    "avx": False,
                    "avx2": False,
                }

                pkg['executables'].append({
                    'file': dest_fn,
                    'label': label,
                    'properties': props,
                })
            elif group.startswith('Win') and fn.endswith('.exe'):
                if 'fred2' in fn:
                    if '-FASTDBG' in fn:
                        label = 'FRED2 Debug'
                    else:
                        label = 'FRED2'

                elif 'qtfred' in fn:
                    if '-FASTDBG' in fn:
                        label = 'QtFRED Debug'
                    else:
                        label = 'QtFRED'

                else:
                    if '-FASTDBG' in fn:
                        label = 'Fast Debug'
                    else:
                        label = None

                props = {
                    "x64": "x64" in fn,
                    "sse2": "SSE2" in fn or "AVX" in fn,  # AVX implies SSE2
                    "avx": "AVX" in fn,  # This conveniently also covers the AVX2 case since AVX2 implies AVX
                    "avx2": "AVX2" in fn,
                }

                pkg['executables'].append({
                    'file': dest_fn,
                    'label': label,
                    'properties': props,
                })

        meta["packages"].append(pkg)

    return meta


@retry_multi(5)
def nebula_request(session, kind, path, **kwargs):
    uri = "https://fsnebula.org/api/1/{}".format(path)

    request_args = {
        "timeout": GLOBAL_TIMEOUT
    }
    request_args.update(kwargs)

    return session.request(kind, uri, **request_args)


def submit_release(meta, config):
    try:
        with requests.Session() as session:
            print('Logging into Nebula...')
            result = nebula_request(session, 'post', 'login', data={
                'user': config['nebula']['user'],
                'password': config['nebula']['password']
            })

            if result.status_code != 200:
                print('Login failed!')
                return False

            data = result.json()
            if not data['result']:
                print('Login failed!')
                return False

            print('Submitting release...')
            result = nebula_request(session, 'post', 'mod/release', headers={
                'X-KN-TOKEN': data['token']
            }, json=meta)

            if result.status_code != 200:
                print('Request failed!')
                return False

            data = result.json()
            if not data['result']:
                print('ERROR: ' + data['reason'])
                return False

            print('Success!')
            return True
    except Exception:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback)
        return False



# Retrives a log of commit messages and attempts to pretty them into a human-readbile format

import argparse
import os
import re
import requests
import sys

from datetime import datetime
from subprocess import check_output

MAX_PAGES = 100  # maximum number of PR pages to go through
warnings = 0

def get_last_page(response: requests.Response) -> int:
    """
    @brief Extract the last page number from the response header

    @param[in] response The response from a GET call

    @returns The page number of the last page, or MAX_PAGES
    """

    # The link header is a dictionary in string form, and for that matter the
    # value is in front of the key.  It is of the form '<<url1>>; rel=<key1>, <url2>, <key2>'
    link = response.headers['link']
    link = link.split(',')
    for x in link:
        y = x.split(';')        # <url>; rel=<key>
        if "last" not in y[1]:  # look for rel='last'
            continue
        # grab the page number from the URL
        # group(0) is `page=123`, group(1) is `123`
        last_page = int( re.search(r'page=(\d+)', y[0]).group(1) )
        if not last_page:
            print("Error: Could not find page number in url!  URL: %s", y[0])
            sys.exit(1)

        if last_page > MAX_PAGES:
            return MAX_PAGES
        else:
            return last_page
    

def main():
    parser = argparse.ArgumentParser(description="Retrieves a log of commit messages and attempts to pretty them into a human-readable format")
    parser.add_argument("start_date",
                        help="Day of the earliest commit to add to the log.")
    parser.add_argument("-u", "--user", nargs=2, metavar=("NAME", "TOKEN"), default=["", ""],
                        help="Github username and token/password")

    args = parser.parse_args()

    os.chdir(os.path.dirname(__file__))	# Change working directory to this file's directory
    start_date = datetime.strptime(args.start_date, "%Y-%m-%d")

    print("start_date: {}".format(start_date))

    # Retrive the git log for master, since start date, order by date, in ascending order, use YYYY-MM-DD date format,
    #   each line is of the form "<author short date> | <message>", only grab commits from master,
    #   output to `fsofordigest.txt`,
    
    # check_output(("git log master", "--after=2021-09-15", "--date-order", "--reverse", "--date=short",
    # "--pretty=\"%as | %s\"",  "--first-parent",
    # "--output fsofordigest.txt"), text=True)

    # The commit messages are either of the form "<body> (#<PR_number>)" or "Merge pull request #<PR_number> from <source_repo>"
    # For our purposes, we want them be all of the form "<body> (#<PR_number>)"
    # Typically, the second form does not have a usable body, so we'll have to pull the PR title from github

    # Github's REST API allows us to retrieve PR's, but due to it being HTTP requests it has to page in the data.
    # We can loop over the GET request and check for a merge date.
    # The GET request doesn't have sorting by merge date, but does have sorting by the last time it was updated... which is usually
    # also the merge date but not always.

    # Key attributes within the PR JSON objects:
    #   "merged_at: <date>"     # Date the PR was merged in
    #   "merged: <bool>"        # If the PR was merged yet
    #   "html_url: <string>"    # URL of the PR's HEAD commit, used by some clients to pull in the PR branch
    #   "number: <int>"         # The PR id number
    #   "title: <string>"       # Title of the PR
    #   "body: <string>"        # The body paragraph detailing the PR
    #   "labels: [label_object]"    # All labels attached to this PR

    changelog = {}
    url = "https://api.github.com/repos/scp-fs2open/fs2open.github.com/pulls"
    page = 1
    last_page = MAX_PAGES
    while page <= last_page:
        # GET <url>?base=master&sort=updated&direction=desc&state=closed&page=1
        response = requests.get(url,
        auth=(args.user[0], args.user[1]),
        params={
            "base": "master",
            "direction": "desc",
            "page": page,
            "state": "closed",
            "sort": "updated"
        })

        # Check if good response
        if (response.status_code != 200) :
            print("Error: ", response.status_code)
            print(response.json()['message'])
            sys.exit(1)
        else:
            print("Request successful: 200")
            print("page={}".format(page))

        # Update the last_page index at first iteration and possibly last iteration
        # This should cover the edge case of a new PR being added to the database before this script is finished
        if (page == 1) or (page == last_page):
            last_page = get_last_page(response)
            print("last_page={}".format(last_page))

        # assume pulls are organized from newest to oldest
        for pull in response.json():
            if (pull['state'] != 'closed'):
                # This pull is not merged, so ignore
                print("Found irrelevant pull, skipping...")
                print("state: {}; number: {}".format(pull['state'], pull['number']))
                continue
            
            if (pull['merged_at'] == None):
                # This pull wasn't merged, skip
                print("Found closed pull #{} ({}), skipping...".format(pull['number'], pull['closed_at']))
                continue

            merge_date = datetime.strptime(pull['merged_at'], "%Y-%m-%dT%H:%M:%SZ")
            updated_date = datetime.strptime(pull['updated_at'], "%Y-%m-%dT%H:%M:%SZ")
            if (merge_date < start_date) and (updated_date >= start_date):
                # This pull is too old, but was updated within our timeframe. skip
                print("Found old pull #{} ({}), skipping...".format(pull['number'], pull['merged_at']))
                continue

            if (updated_date < start_date):
                # This pull is too old
                print("Found old pull #{} ({}), stopping loop...".format(pull['number'], pull['updated_at']))
                page = last_page + 1
                break

            # Add the pull to the changelog, indexed by its pull.number
            if pull['number'] not in changelog:
                # only add if the key isn't there already
                changelog[pull['number']] = {
                    "date":     merge_date,
                    "title":    pull['title'],
                    "labels":   pull['labels'],
                    "url":      pull['html_url']
                }
                print("Added pull #{} ({})".format(pull['number'], merge_date))
            # Else, silently ignore duplicate (for now...)

        page += 1 # iterate to next page


    # Write the changelog to file
    # 2021-09-28 | (#3661) Fix homing child weapons in niche circumstances 
    print("Writing to change.log...")
    str = ''
    for key, value in changelog.items():
        str += "{} | (#{}) {} \n".format(value['date'].strftime("%Y-%m-%d"), key, value['title'])
    
    with open("change.log", "w") as file:
        file.write(str)

    print("Done!")

main()


# Post-build python script
# Uploads build packages to Nebula and creates a forum post on https://www.hard-light.net
# Uses the following environment variables (Set by the github action workflow before execution)

# FSO versioning
# FSO_VERSION_MAJOR
# FSO_VERSION_MINOR
# FSO_VERSION_BUILD

# Credentials into IndieGames file host
# INDIEGAMES_USER:
# INDIEGAMES_SSHPASS:

# Credentials into Nebula Mod Database
# NEBULA_USER:
# NEBULA_PASSWORD:

# Credentials into hard-light.net for forum post
# HLP_API:
# HLP_KEY:

# LINUX_RESULT: bool, if the linux builds were successfully uploaded
# WINDOWS_RESULT: bool, if the windows builds were successfully uploaded

import sys
import os
import re
from subprocess import check_output
from datetime import datetime
from itertools import groupby

import semantic_version

# local .py files
import file_list
import installer
import nebula
import forum

# Compile regexes for extracting version components
MAJOR_VERSION_PATTERN = re.compile("(?:set_if_not_defined|set)\(FSO_VERSION_MAJOR (\d+)\)")
MINOR_VERSION_PATTERN = re.compile("(?:set_if_not_defined|set)\(FSO_VERSION_MINOR (\d+)\)")
BUILD_VERSION_PATTERN = re.compile("(?:set_if_not_defined|set)\(FSO_VERSION_BUILD (\d+)\)")
REVISION_VERSION_PATTERN = re.compile("(?:set_if_not_defined|set)\(FSO_VERSION_REVISION (\d+)\)")
REVISION_STR_VERSION_PATTERN = re.compile("(?:set_if_not_defined|set)\(FSO_VERSION_REVISION_STR (\w*)\)")

LOG_FORMAT = """
------------------------------------------------------------------------%n
commit %h%nAuthor: %an <%ad>%n
Commit: %cn <%cd>%n%n    %s
"""
DATEFORMAT_VERSION = "%Y%m%d"
DATEFORMAT_FORUM = "%d %B %Y"

def _match_version_number(text, regex):
	"""! Extracts the version component represented by `regex` from `text`
	
	@param [in] `text`  Filename with version string 'MAJOR.MINOR.BUILD'
	@param [in] `regex` Regex pattern of component to look for

	@return The version component as integer
	"""
	match = regex.search(text)
	return int(match.group(1))

def _match_version_str(text, regex):
	"""! Extracts the version component represented by `regex` from `text`
	
	@param [in] `text`  Filename with version string 'MAJOR.MINOR.BUILD'
	@param [in] `regex` Regex pattern of component to look for

	@return The version component as integer
	"""
	match = regex.search(text)
	return match.group(1)
	

def get_source_version(date_version: datetime, tag_name: str) -> semantic_version.Version:
	"""! Retrieves the version string from version_override.cmake (and version.cmake)

	@param[in] `date_version` Date this build is published for release
	@param[in] `tag_name`     Tag of this release, used to determine if release, rc, or nightly

	@return If release: `FSO_VERSION_MAJOR.FSO_VERSION_MINOR.FSO_VERSION_BUILD`
	@return If rc:      `FSO_VERSION_MAJOR.FSO_VERSION_MINOR.FSO_VERSION_BUILD-FSO_VERSION_REVISION_STR`
	@return If nightly: `FSO_VERSION_MAJOR.FSO_VERSION_MINOR.FSO_VERSION_BUILD-FSO_VERSION_REVISION_STR` (uses both version_override.cmake and version.cmake)

	@note This is the version used to identify in Nebula and the forums.  version_override.cmake is used for the builds
	  made by CI tools and version.cmake is used for builds made by devs.
	"""

	with open(os.path.join("..", "..", "version_override.cmake"), "r") as f:
		filetext = f.read()
		major = _match_version_number(filetext, MAJOR_VERSION_PATTERN)
		minor = _match_version_number(filetext, MINOR_VERSION_PATTERN)
		build = _match_version_number(filetext, BUILD_VERSION_PATTERN)
		# revision = _match_version_number(filetext, REVISION_VERSION_PATTERN)
		revision_str = _match_version_str(filetext, REVISION_STR_VERSION_PATTERN)

	print("Parsing version_override.cmake...")
	if "rc" in tag_name.lower():
		version = semantic_version.Version("{}.{}.{}-{}".format(major, minor, build, revision_str))

	elif "release" in tag_name.lower():
		version = semantic_version.Version("{}.{}.{}".format(major, minor, build))

	elif "nightly" in tag_name.lower():
		# Nightly's version_override.cmake only has FSO_VERSION_REVISION and FSO_VERSION_REVISION_STR set
		# So, we have to pull the rest of the version string from version.cmake
		print("Parsing version.cmake...")
		with open(os.path.join("..", "..", "cmake", "version.cmake"), "r") as f:
			major = _match_version_number(filetext, MAJOR_VERSION_PATTERN)
			minor = _match_version_number(filetext, MINOR_VERSION_PATTERN)
			build = _match_version_number(filetext, BUILD_VERSION_PATTERN)
		version = semantic_version.Version("{}.{}.{}-{}".format(major, minor, build, revision_str))
	
	else:	
		print("  ERROR: malformed tag_name %s" % tag_name)
		sys.exit(1)

	print("  version: {}".format(version))
	print("  version.prerelease: {}".format(version.prerelease))
	return version


def main():
	os.chdir(os.path.dirname(__file__))	# Change working directory to this file's directory

	# Aggregate configuration data in a dictionary
	config = {
		"github": {
			"repo": os.environ["GITHUB_REPO"],
		},
		"ftp": {
			"host": "scp.indiegames.us",
			"user": os.environ["INDIEGAMES_USER"],
			"pass": os.environ["INDIEGAMES_SSHPASS"],
			"path": "public_html/builds/{type}/{version}/",
			"mirrors": (
				"https://perses.feralhosting.com/datacorder/builds/{type}/{version}/{file}",
				"https://scp.indiegames.us/builds/{type}/{version}/{file}",
			),
		},
		"nebula": {
			"user": os.environ["NEBULA_USER"],
			"password": os.environ["NEBULA_PASSWORD"],
		},
		"hlp": {
			"api": os.environ["HLP_API"],
			"key": os.environ["HLP_KEY"],
		},
		"templates": {
			"nightly": "nightly.mako",
			"release": "release.mako",
		},
		"nightly": {
			"hlp_board": 173,
		},
		"release": {
			"hlp_board": 50,
		},
	}

	# bail if this script is run from a workflow that doesn't identify as a release or nightly
	if sys.argv[1] not in ("release", "nightly"):
		print("ERROR: Invalid release mode %s passed. Expected release or nightly!" % sys.argv[1])
		sys.exit(1)

	linux_success = os.environ["LINUX_RESULT"] == "success"
	windows_success = os.environ["WINDOWS_RESULT"] == "success"
	if (sys.argv[1] == "release"):
		if not (linux_success and windows_success):
			# bail if this is a release and either build workflow failed
			print("ERROR: One or more builds failed to publish! Cannot post release.")
			sys.exit(1)
	else:
		if not (linux_success or windows_success):
			# bail if this is a nightly build and both build workflows failed
			print("ERROR: All builds failed to publish! Cannot post nightly.")
			sys.exit(1)
	
	# bail if tag_name doesn't match with sys.argv[1]
	tag_name = os.environ["RELEASE_TAG"]	##!< commit tag string
	if (sys.argv[1] == "release") and (not tag_name.startswith("release_")):
		print("ERROR: Invalid tag name detected. Expected release_ prefix for release builds but found %s (full ref = %s)" % tag_name, os.environ["GITHUB_REF"])
		sys.exit(1)

	if (sys.argv[1] == "nightly") and (not tag_name.startswith("nightly_")):
		print("ERROR: Invalid tag name detected. Expected nightly_ prefix for nightly builds but found %s (full ref = %s)" % tag_name, os.environ["GITHUB_REF"])
		sys.exit(1)

	# bail if tag_name does not exist in the repo
	if not check_output(("git", "show-ref", "--tags", "refs/tags/%s" % tag_name), text=True):
		print("ERROR: Couldn't find tag %s in repo (git show-ref)!" % tag_name)
		sys.exit(1)

	# Populate version and date
	date = datetime.now()	##!< current date
	version = get_source_version(date.strftime(DATEFORMAT_VERSION), tag_name)	##!< form full version string
	
	# find the previous tag release (used by nightly)
	tags = check_output(("git", "for-each-ref", "--sort=-taggerdate", "--format='%(tag)'", "refs/tags"), text=True).splitlines()	# retrieve all tags in the repo by using git on the shell
	# NOTE: check_output returns canonical string representation. use repr() on strings its being compared to
	previous_tag = None
	found = False
	for tag in tags:
		if found:
			# Look for the "previous" tag with the same configuration ('release_' or 'nightly_')
			# Assumes older tags have a higher index than newer
			if tag.startswith("\'" + sys.argv[1] + "_"):
				previous_tag = tag
				break
		elif tag == repr(tag_name):
			# Ok, found the tag
			found = True

	if not found:
		# couldn't find tag, bail
		print("ERROR: Couldn't find tag %s in repo! (git for-each-ref)" % tag_name)
		sys.exit(1)

	if (sys.argv[1] == "nightly") and (not previous_tag):
		# couldn't find previous tag, bail
		print("ERROR: Could not find a previous tag for %s!" % tag_name)
		sys.exit(1)

	if sys.argv[1] == "release":
		# Release specific action
		print("Retrieving file list...")
		files, sources = file_list.get_release_files(tag_name, config)

		print("Hashing files...")
		for file in files:
			installer.get_hashed_file_list(file)

		# Construct the file groups
		# split the file list into a dictionary of {group: files} pairs using `groupby()`
		# Since `groupby()` returns a iterator reference to the original files list, we have to do a weird hand toss
		# to save it as a dictionary
		## NOTE z64: files need to be sorted or else key duplication will occur!
		groups = {key: file_list.FileGroup(key, list(group)) for key, group in groupby(files, lambda g: g.group)}
		
		# Error check for all group keys and subkeys to be used in the mako.  Do this here for better debugging QoL
		if "Win32" in groups.keys():
			if "AVX" not in groups["Win32"].subFiles.keys():
				print("ERROR: No Win32-AVX builds were detected!")
				sys.exit(1)
		else:
			print("ERROR: No Win32 builds were detected!")
			sys.exit(1)

		if "Win64" in groups.keys():
			if "AVX" not in groups["Win64"].subFiles.keys():
				print("ERROR: No x64-AVX builds were detected!")
				sys.exit(1)
		else:
			print("ERROR: No x64 builds were detected!")
			sys.exit(1)
		
		if "Linux" not in groups.keys():
			print("ERROR: No Linux builds were detected!")
			sys.exit(1)

		if "Mac" not in groups.keys():
			print("ERROR: No Mac builds were detected!")
			sys.exit(1)
		
		# z64: What dose this do???
		# print(installer.render_installer_config(version, groups, config))

		# Publish release builds to Nebula
		print("Publishing to Nebula...")
		nebula.submit_release(
			nebula.render_nebula_release(version, "rc" if version.prerelease else "stable", files, config),
			config)

		# Publish forum post, using the release.mako template
		print("Publishing to Hard-light.net...")
		fapi = forum.ForumAPI(config)
		fapi.post_release(date.strftime(DATEFORMAT_FORUM), version, groups, sources)

	else:
		# Dead for now, nightly publish is done elsewhere
		# Nightly specific action
		print("Retrieving file list...")
		files = file_list.get_ftp_files("nightly", tag_name, config)

		print("Hashing files...")
		for file in files:
			installer.get_hashed_file_list(file)

		print("Publishing to Nebula...")
		nebula.submit_release(
			nebula.render_nebula_release(version, "nightly", files, config),
			config)

		# Compile commit messages into a log with all commits between this nightly and the previous nightly
		log = check_output(("git", "log", "%s^..%s^" % (previous_tag, tag_name), "--no-merges", "--stat", "--pretty=format:\"%s\"" % LOG_FORMAT.strip()))

		# Publish forum post, using the nighty.mako template
		print("Publishing to Hard-light.net...")
		fapi = forum.ForumAPI(config)
		fapi.post_nightly(date.strftime(DATEFORMAT_FORUM), os.environ["GITHUB_SHA"], files, log, (linux_success and windows_success))

	print("Done!")


if __name__ == "__main__":
	main()



import os
import hashlib
import tarfile
import traceback
import zipfile
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

import requests
import sys
from mako.template import Template

from file_list import ReleaseFile
from util import GLOBAL_TIMEOUT


def _download_file(url, dest_file, session):
    print("Downloading " + url)
    # NOTE the stream=True parameter
    r = session.get(url, stream=True, timeout=GLOBAL_TIMEOUT)
    for chunk in r.iter_content(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            dest_file.write(chunk)
    dest_file.flush()


def _gen_hash(fileobj, hash_alg):
    h = hashlib.new(hash_alg)
    while True:
        data = fileobj.read(4096)

        if not data:
            break

        h.update(data)

    return h.hexdigest()


def get_hashed_file_list(file: ReleaseFile, session: requests.Session = None, hash_alg: str = "sha256"):
    if session is None:
        session = requests.Session()

    with NamedTemporaryFile('w+b', suffix=file.filename) as local_file:
        _download_file(file.url, local_file, session)

        filename = local_file.name
        hash_list = []

        local_file.seek(0)
        file.hash = _gen_hash(local_file, hash_alg)
        file.size = os.stat(filename).st_size

        try:
            if tarfile.is_tarfile(filename):
                with tarfile.open(filename) as archive:
                    for entry in archive:
                        if not entry.isfile():
                            continue

                        print("Computing hash for " + entry.name)
                        fileobj = archive.extractfile(entry)
                        hash_list.append((entry.path, _gen_hash(fileobj, hash_alg)))
            elif zipfile.is_zipfile(filename):
                with ZipFile(filename) as archive:
                    for entry in archive.infolist():
                        # Python 3.6 has is_dir but that version is relatively new so we'll use the "safe" version here
                        if entry.filename.endswith('/'):
                            continue

                        print("Computing hash for " + entry.filename)
                        with archive.open(entry) as fileobj:
                            hash_list.append((entry.filename, _gen_hash(fileobj, hash_alg)))
            else:
                raise NotImplementedError("Unsupported archive type!")
        except:
            traceback.print_exception(*sys.exc_info())
            return

        file.content_hashes = hash_list


def render_installer_config(version, groups, config):
    template = Template(filename=config["templates"]["installer"], module_directory='/tmp/mako_modules')
    return template.render(**{
        "version": version,
        "groups": groups,
    }).strip("\n")



import re # regex module
from ftplib import FTP, error_perm
from itertools import groupby
from typing import List, Tuple, Dict

import requests  # HTTP requests module

from util import retry_multi, GLOBAL_TIMEOUT	# from util.py


class ReleaseFile:
    """! Class representing a Released file on Nebula

    `name`: str
        Mod (or build) name,
    `url`: str
        Primary host URL,
    `group`: str
        Mod group string,
    `subgroup`: str
        Mod subgroup string,
    `mirrors`: List[str]
        List of URL's of FTP mirrors
    """
    def __init__(self, name, url, group, subgroup=None, mirrors=None):
        if mirrors is None:
            mirrors = []
        self.mirrors = mirrors
        self.subgroup = subgroup
        self.group = group
        self.url = url
        self.name = name

        self.base_url = "/".join(url.split('/')[0:-1]) + "/"
        self.filename = url.split('/')[-1]

        # A list of tuples of (filename, hash)
        self.content_hashes = None

        self.hash = None
        self.size = 0
    
    def __repr__(self):
        return repr((self.name))


class SourceFile:
    """! Class represeting a source file

    `name`: str
        File name,
    `url`: str
        FTP URL,
    `group`
        <unknown>
 
    @details More details
    """
    def __init__(self, name, url, group):
        self.group = group
        self.url = url
        self.name = name


class FileGroup:
    """! Represents a file group

    `name`: str
        Name of this group
    `files`: List[ReleaseFile]
        List of files within this group
    `mainFile`: str
        If this FileGroup has a subgroup, `mainFile` is the head of that group
    `subFiles`: List[ReleaseFile]
        Files within a subgroup
    """

    def __init__(self, name, files: List[ReleaseFile]):
        self.files = files
        self.name = name

        if len(files) == 1:
            self.mainFile = files[0]
            self.subFiles = {}
        else:
            self.mainFile = None
            subFiles = []
            for file in files:
                # We only have subcategories for Windows where SSE2 is the main group
                if file.subgroup == "SSE2":
                    self.mainFile = file
                else:
                    subFiles.append(file)

            self.subFiles = dict(((x[0], next(x[1])) for x in groupby(subFiles, lambda f: f.subgroup)))


def get_release_files(tag_name, config) -> Tuple[List[ReleaseFile], Dict[str, SourceFile]]:
    """! Brief Gets the binary and source files from the Github Release server

    @param[in] `tag_name` Git tag of the current release
    @param[in] `config`   confi metadata set in main.py

    @returns `List[ReleaseFile]`        List of release files
    @returns `Dict[str, SourceFile]`    Dictionary of source files

    @details Sends an `HTTP GET` request to github using their REST API to retrieve metadata.  The files are not
        actually downloaded here, just their metadata is gathered and organized in their respective container for later
        use.
    """

    @retry_multi(5)	# retry at most 5 times
    def execute_request(path):
        """!
        @brief Performs a GET request with the given path. To be used with Github's REST API.
        @returns If successful, returns a .JSON object
        """
        headers = {
            "Accept": "application/vnd.github.v3+json"
        }
        url = "https://api.github.com" + path

        # GET https://api.github.com/<path> Accept: "application/vnd.github.v3+json"

        response = requests.get(url, headers=headers, timeout=GLOBAL_TIMEOUT)

        response.raise_for_status() # Raise a RequestException if we failed, and trigger retry

        return response.json()

    build_group_regex = re.compile("fs2_open_.*-builds-([^.-]*)(-([^.]*))?.*")  # regex for matching binary .zip's and .7z's
    source_file_regex = re.compile("fs2_open_.*-source-([^.]*)?.*") # regex for matching source .zip's and .7z's

    # Get the github release metadata of the given tag name
    response = execute_request(
        "/repos/{}/releases/tags/{}".format(config["github"]["repo"], tag_name))

    # Extract the binary and source files from the response["asset"] metadata
    binary_files = []
    source_files = {}
    for asset in response["assets"]:
        url = asset["browser_download_url"]
        name = asset["name"]

        group_match = build_group_regex.match(name)

        if group_match is not None:
            platform = group_match.group(1)
            # x64 is the Visual Studio name but for consistency we need Win64
            if platform == "x64":
                platform = "Win64"

            binary_files.append(ReleaseFile(name, url, platform, group_match.group(3)))
        else:
            group_match = source_file_regex.match(name)

            if group_match is None:
                continue

            group = group_match.group(1)

            source_files[group] = SourceFile(name, url, group)

    binary_files.sort(key=lambda ReleaseFile: ReleaseFile.name)

    return binary_files, source_files


def get_ftp_files(build_type, tag_name, config) -> List[ReleaseFile] :
    """!
    @brief Gets file metadata for nightlies hosted on FTP, as determined by config["ftp"] attributes
    
    @param [in] `build_type` Unknown str
    @param [in] `tag_name`   Github tag name of the release
    @param [in] `config`     config metadata set in main.py
    """

    tag_regex = re.compile("nightly_(.*)")
    build_group_regex = re.compile("nightly_.*-builds-([^.]+).*")

    files = []
    try:
        with FTP(config["ftp"]["host"], config["ftp"]["user"], config["ftp"]["pass"]) as ftp:
            # extract version
            version_str = tag_regex.match(tag_name).group(1)

            # extract filepath w/ version
            # then list all ftp hits with that path
            path_template = config["ftp"]["path"]
            path = path_template.format(type=build_type, version=version_str)
            file_entries = list(ftp.mlsd(path, ["type"]))

            # get all ftp hits of type file
            for entry in file_entries:
                if entry[1]["type"] == "file":
                    files.append(entry[0])
    except error_perm:
        print("Received permanent FTP error!")
        return []

    out_data = []
    for file in files:
        # from the file list, extract only nightly files
        file_match = build_group_regex.match(file)
        if file_match is None:
            print("Ignoring non nightly file '{}'".format(file))
            continue

        group_match = file_match.group(1)
        primary_url = None
        mirrors = []

        # x64 is the name Visual Studio uses but Win64 works better for us since that gets displayed in the nightly post
        if "x64" in group_match:
            group_match = group_match.replace("x64", "Win64")

        # construct the download URL list for all mirrors.  The first listed ftp location is taken as the Primary
        for mirror in config["ftp"]["mirrors"]:
            download_url = mirror.format(type=build_type, version=version_str, file=file)
            if primary_url is None:
                primary_url = download_url
            else:
                mirrors.append(download_url)

        # Form the List[ReleaseFile] list with the download URL links
        out_data.append(ReleaseFile(file, primary_url, group_match, None, mirrors))

    return out_data


from typing import List

import requests
import semantic_version
from mako.template import Template

from file_list import ReleaseFile


class ForumAPI:
    POST_SIZE_MAX = 50000
    """!
    Maximum number of characters allowed in a post, including BBCode (03/17/2022)

    Check with HLP admins occasionally to verify this limit, they can set it in the forum's admin panel
    """

    def __init__(self, config):
        self.config = config

    def create_post(self, title, content, board):
        """!
        @brief Create a post on the forums

        @param[in] title    Title of the post
        @param[in] content  Body of the post
        @param[in] board    Which board/forum to post to

        @returns { "thread_url": <url_of_created_thread> }
        """

        if (self.config["hlp"]["api"] == "") or (self.config["hlp"]["key"] == ""):
            print("Post failed! No API or API_KEY given!")
            return

        resp = requests.post(self.config["hlp"]["api"], data={
            "api_key": self.config["hlp"]["key"],
            "board": str(board),
            "subject": title,
            "body": content
        })

        if resp.status_code != 200:
            print("Post failed! Response: %s" %resp.text)
        
        return resp.json()

    def post_nightly(self, date, revision, files, log, success):
        """!
        @brief Post a new Nightly topic on the forums using nightly.mako

        @param[in] date         The date to use
        @param[in] version      The build's version
        @param[in] groups       dict[Any, file_list.FileGroup]; FileGroups of SSE2, AVX, etc.
        @param[in] sources      URL's of builds on github or fs2net
        """

        print("Posting nightly thread...")

        title = "Nightly: {} - Revision {}".format(date, revision)

        template = Template(filename=self.config["templates"]["nightly"])
        rendered = template.render_unicode(**{
            "date": date,
            "revision": revision,
            "files": files,
            "log": log,
            "success": success
        })

        print("Creating post...")
        self.create_post(title, rendered, self.config["nightly"]["hlp_board"])

    def post_release(self, date, version: semantic_version.Version, groups, sources):
        """!
        @brief Post a new Release topic on the forums using release.mako.

        @param[in] date         The date to use
        @param[in] version      The build's version
        @param[in] groups       dict[Any, file_list.FileGroup]; FileGroups of SSE2, AVX, etc.
        @param[in] sources      URL's of builds on github or fs2net

        @returns If Successful, the URL of the created post
        """

        print("Posting release thread...")

        title = "Release: {}".format(version)

        template = Template(
            filename=self.config["templates"]["release"].format(major=version.major, minor=version.minor),
            module_directory='/tmp/mako_modules')
        rendered = template.render_unicode(**{
            "date": date,
            "version": version,
            "groups": groups,
            "sources": sources
        }).strip("\n")

        print("Creating post...")

        json = self.create_post(title, rendered, self.config["release"]["hlp_board"])
        return json["thread_url"]



# Class for use with interfacing with github via HTTP requests

import requests

from util import retry_multi, GLOBAL_TIMEOUT

class GithubAPI:
    headers = {
        "Accept": "application/vnd.github.v3+json"
    }
    url_root = "https://api.github.com"

    def __init__(self, config):
        self.config = config

    @retry_multi(5)
    def get(self, path):
        """!
        @brief Performs a GET request with the given path. To be used with Github's REST API.
        @returns If successful, returns a .JSON object
        """
        url = self.url_root + path

        # GET https://api.github.com/<path> Accept: "application/vnd.github.v3+json"

        response = requests.get(url, headers=self.headers, timeout=GLOBAL_TIMEOUT)

        response.raise_for_status() # Raise a RequestException if we failed, and trigger retry

        return response.json()
    
    def get_release(self, tag_name):
        """!
        @brief Retrieves the Github Release metadata as a JSON object
        @see https://docs.github.com/en/rest/reference/releases#get-a-release-by-tag-name
        """
        return self.get("/repos/{}/releases/tags/{}".format(self.config["github"]["repo"], tag_name))


    def patch(self, path, data):
        """!
        @brief Performs a PATCH request with the given path and body data (as dictionary)
        @returns If successful, returns a .JSON object
        """
        url = self.url_root + path

        response = requests.patch(url, headers=self.headers, timeout=GLOBAL_TIMEOUT, data=data)

        response.raise_for_status()

        return response.json()
        
    def update_release(self, release_id, tag_name="", name="", body="", draft=None, prerelease=None):
        """!
        @brief Updates a Github Release
        @param[in] release_id   The id of the release, as obtained by get_release()
        @param[in] tag_name     (optional) The git tag this release is associated with
        @param[in] name         (optional) The name of the release
        @param[in] body         (optional) Body text describing the release
        @param[in] draft        (optional) If true, the release is marked as a "draft"
        @param[in] prerelease   (optional) If true, the release is marked as a "prerelease"

        @note If an optional argument is not provided, then its associated data is not changed
        """
        path = "/repos/{}/releases/{}".format(self.config["github"]["repo"], release_id)

        # Add to the data JSON if options are specified
        data = {}
        if tag_name != "":
            data["tag_name"] = tag_name
        
        if name != "":
            data["name"] = name

        if body != "":
            data["body"] = body
        
        if draft != None:
            data["draft"] = data

        if prerelease != None:
            data["prerelease"] = prerelease
        
        self.patch(path, data)




#!/usr/bin/env python3
#
#===- clang-tidy-diff.py - ClangTidy Diff Checker ------------*- python -*--===#
#
# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
#===------------------------------------------------------------------------===#

r"""
ClangTidy Diff Checker
======================

This script reads input from a unified diff, runs clang-tidy on all changed
files and outputs clang-tidy warnings in changed lines only. This is useful to
detect clang-tidy regressions in the lines touched by a specific patch.
Example usage for git/svn users:

  git diff -U0 HEAD^ | clang-tidy-diff.py -p1
  svn diff --diff-cmd=diff -x-U0 | \
      clang-tidy-diff.py -fix -checks=-*,modernize-use-override

"""

import argparse
import glob
import json
import multiprocessing
import os
import re
import shutil
import subprocess
import sys
import tempfile
import threading
import traceback

try:
    import yaml
except ImportError:
    yaml = None

is_py2 = sys.version[0] == '2'

if is_py2:
    import Queue as queue
else:
    import queue as queue


def run_tidy(task_queue, lock, timeout):
    watchdog = None
    while True:
        command = task_queue.get()
        try:
            proc = subprocess.Popen(command,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)

            if timeout is not None:
                watchdog = threading.Timer(timeout, proc.kill)
                watchdog.start()

            stdout, stderr = proc.communicate()

            with lock:
                sys.stdout.write(stdout.decode('utf-8'))
                sys.stdout.flush()
                if stderr:
                    sys.stderr.write(stderr.decode('utf-8'))
                    sys.stderr.flush()
        except Exception as e:
            with lock:
                sys.stderr.write('Failed: ' + str(e) + ': '.join(command) + '\n')
        finally:
            with lock:
                if (not timeout is None) and (not watchdog is None):
                    if not watchdog.is_alive():
                        sys.stderr.write('Terminated by timeout: ' +
                                         ' '.join(command) + '\n')
                    watchdog.cancel()
            task_queue.task_done()


def start_workers(max_tasks, tidy_caller, task_queue, lock, timeout):
    for _ in range(max_tasks):
        t = threading.Thread(target=tidy_caller, args=(task_queue, lock, timeout))
        t.daemon = True
        t.start()

def merge_replacement_files(tmpdir, mergefile):
    """Merge all replacement files in a directory into a single file"""
    # The fixes suggested by clang-tidy >= 4.0.0 are given under
    # the top level key 'Diagnostics' in the output yaml files
    mergekey = "Diagnostics"
    merged = []
    for replacefile in glob.iglob(os.path.join(tmpdir, '*.yaml')):
        content = yaml.safe_load(open(replacefile, 'r'))
        if not content:
            continue # Skip empty files.
        merged.extend(content.get(mergekey, []))

    if merged:
        # MainSourceFile: The key is required by the definition inside
        # include/clang/Tooling/ReplacementsYaml.h, but the value
        # is actually never used inside clang-apply-replacements,
        # so we set it to '' here.
        output = { 'MainSourceFile': '', mergekey: merged }
        with open(mergefile, 'w') as out:
            yaml.safe_dump(output, out)
    else:
        # Empty the file:
        open(mergefile, 'w').close()


def main():
    parser = argparse.ArgumentParser(description=
                                     'Run clang-tidy against changed files, and '
                                     'output diagnostics only for modified '
                                     'lines.')
    parser.add_argument('-clang-tidy-binary', metavar='PATH',
                        default='clang-tidy',
                        help='path to clang-tidy binary')
    parser.add_argument('-p', metavar='NUM', default=0,
                        help='strip the smallest prefix containing P slashes')
    parser.add_argument('-regex', metavar='PATTERN', default=None,
                        help='custom pattern selecting file paths to check '
                             '(case sensitive, overrides -iregex)')
    parser.add_argument('-iregex', metavar='PATTERN', default=
    r'.*\.(cpp|cc|c\+\+|cxx|c|cl|h|hpp|m|mm|inc)',
                        help='custom pattern selecting file paths to check '
                             '(case insensitive, overridden by -regex)')
    parser.add_argument('-j', type=int, default=1,
                        help='number of tidy instances to be run in parallel.')
    parser.add_argument('-timeout', type=int, default=None,
                        help='timeout per each file in seconds.')
    parser.add_argument('-fix', action='store_true', default=False,
                        help='apply suggested fixes')
    parser.add_argument('-checks',
                        help='checks filter, when not specified, use clang-tidy '
                             'default',
                        default='')
    parser.add_argument('-path', dest='build_path',
                        help='Path used to read a compile command database.')
    if yaml:
        parser.add_argument('-export-fixes', metavar='FILE', dest='export_fixes',
                            help='Create a yaml file to store suggested fixes in, '
                                 'which can be applied with clang-apply-replacements.')
    parser.add_argument('-extra-arg', dest='extra_arg',
                        action='append', default=[],
                        help='Additional argument to append to the compiler '
                             'command line.')
    parser.add_argument('-extra-arg-before', dest='extra_arg_before',
                        action='append', default=[],
                        help='Additional argument to prepend to the compiler '
                             'command line.')
    parser.add_argument('-quiet', action='store_true', default=False,
                        help='Run clang-tidy in quiet mode')
    clang_tidy_args = []
    argv = sys.argv[1:]
    if '--' in argv:
        clang_tidy_args.extend(argv[argv.index('--'):])
        argv = argv[:argv.index('--')]

    args = parser.parse_args(argv)

    # Extract changed lines for each file.
    filename = None
    lines_by_file = {}
    for line in sys.stdin:
        match = re.search('^\+\+\+\ \"?(.*?/){%s}([^ \t\n\"]*)' % args.p, line)
        if match:
            filename = match.group(2)
        if filename is None:
            continue

        if args.regex is not None:
            if not re.match('^%s$' % args.regex, filename):
                continue
        else:
            if not re.match('^%s$' % args.iregex, filename, re.IGNORECASE):
                continue

        match = re.search('^@@.*\+(\d+)(,(\d+))?', line)
        if match:
            start_line = int(match.group(1))
            line_count = 1
            if match.group(3):
                line_count = int(match.group(3))
            if line_count == 0:
                continue
            end_line = start_line + line_count - 1
            lines_by_file.setdefault(filename, []).append([start_line, end_line])

    if not any(lines_by_file):
        print("No relevant changes found.")
        sys.exit(0)

    max_task_count = args.j
    if max_task_count == 0:
        max_task_count = multiprocessing.cpu_count()
    max_task_count = min(len(lines_by_file), max_task_count)

    tmpdir = None
    if yaml and args.export_fixes:
        tmpdir = tempfile.mkdtemp()

    # Tasks for clang-tidy.
    task_queue = queue.Queue(max_task_count)
    # A lock for console output.
    lock = threading.Lock()

    # Run a pool of clang-tidy workers.
    start_workers(max_task_count, run_tidy, task_queue, lock, args.timeout)

    # Form the common args list.
    common_clang_tidy_args = []
    if args.fix:
        common_clang_tidy_args.append('-fix')
    if args.checks != '':
        common_clang_tidy_args.append('-checks=' + args.checks)
    if args.quiet:
        common_clang_tidy_args.append('-quiet')
    if args.build_path is not None:
        common_clang_tidy_args.append('-p=%s' % args.build_path)
    for arg in args.extra_arg:
        common_clang_tidy_args.append('-extra-arg=%s' % arg)
    for arg in args.extra_arg_before:
        common_clang_tidy_args.append('-extra-arg-before=%s' % arg)

    for name in lines_by_file:
        line_filter_json = json.dumps(
            [{"name": name, "lines": lines_by_file[name]}],
            separators=(',', ':'))

        # Run clang-tidy on files containing changes.
        command = [args.clang_tidy_binary]
        command.append('-line-filter=' + line_filter_json)
        if yaml and args.export_fixes:
            # Get a temporary file. We immediately close the handle so clang-tidy can
            # overwrite it.
            (handle, tmp_name) = tempfile.mkstemp(suffix='.yaml', dir=tmpdir)
            os.close(handle)
            command.append('-export-fixes=' + tmp_name)
        command.extend(common_clang_tidy_args)
        command.append(name)
        command.extend(clang_tidy_args)

        task_queue.put(command)

    # Wait for all threads to be done.
    task_queue.join()

    if yaml and args.export_fixes:
        print('Writing fixes to ' + args.export_fixes + ' ...')
        try:
            merge_replacement_files(tmpdir, args.export_fixes)
        except:
            sys.stderr.write('Error exporting fixes.\n')
            traceback.print_exc()

    if tmpdir:
        shutil.rmtree(tmpdir)


if __name__ == '__main__':
    main()



# miscellaneous utilities

import os
import time
from functools import wraps

from requests import RequestException

GLOBAL_TIMEOUT = 30

def retry_multi(max_retries):
    """!
    @brief Decorator - Retry a function `max_retries` times.
    
    @param [in] `max_retries` Maximum number of times to retry the decorated function before giving up

    @details  Intended for use with functions from the `request` module, only catches `RequestException`'s
    @note This has been copied from https://stackoverflow.com/a/23892489
    """

    def retry(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            num_retries = 0
            while num_retries <= max_retries:
                try:
                    ret = func(*args, **kwargs)
                    break
                except RequestException:
                    if num_retries == max_retries:
                        raise
                    num_retries += 1
                    time.sleep(5)
            return ret

        return wrapper

    return retry


def expand_config_vars(config):
    """
    @brief Expands the shell variable for the repo string from `config`
    """

    config["git"]["repo"] = os.path.expandvars(config["git"]["repo"])
